#include "mpid.h"
#include "mpiddev.h"
#include "mpimem.h"
#include "reqalloc.h"

/*
   Nonblocking, eager send/recv.
   These are ALWAYS for long messages.  Short messages are always
   handled in blocking eager mode.  

   We COULD write the eager code to use nonblocking receives as well
   as sends, but that is too much like the rendezvous code.  Instead,
   this code RECEIVES using the eager blocking approach but SENDS with
   nonblocking sends.
 */

/* Prototype definitions */
int MPID_CH_Eagern_isend ANSI_ARGS(( void *, int, int, int, int, int, int,
				     MPIR_SHANDLE * ));
int MPID_CH_Eagern_cancel_send ANSI_ARGS(( MPIR_SHANDLE * ));
int MPID_CH_Eagern_wait_send ANSI_ARGS(( MPIR_SHANDLE * ));
int MPID_CH_Eagern_test_send ANSI_ARGS(( MPIR_SHANDLE * ));
void MPID_CH_Eagern_delete ANSI_ARGS(( MPID_Protocol * ));

/* 
 * Blocking operations come from chbeager.c
 */
extern int MPID_CH_Eagerb_send ANSI_ARGS(( void *, int, int, int, int, 
					   int, int ));
extern int MPID_CH_Eagerb_recv ANSI_ARGS(( MPIR_RHANDLE *, int, void * ));
extern int MPID_CH_Eagerb_irecv ANSI_ARGS(( MPIR_RHANDLE *, int, void * ));
extern int MPID_CH_Eagerb_save ANSI_ARGS(( MPIR_RHANDLE *, int, void * ));
extern int MPID_CH_Eagerb_unxrecv_start ANSI_ARGS(( MPIR_RHANDLE *, void * ));

/*
 * Definitions of the actual functions
 */

int MPID_CH_Eagern_isend( buf, len, src_lrank, tag, context_id, dest,
			 msgrep, shandle )
void          *buf;
int           len, tag, context_id, src_lrank, dest;
MPID_Msgrep_t msgrep;
MPIR_SHANDLE  *shandle;
{
    int              pkt_len;
    MPID_PKT_LONG_T  pkt;
    
    pkt.mode	   = MPID_PKT_LONG;
    pkt_len	   = sizeof(MPID_PKT_LONG_T); 
    pkt.context_id = context_id;
    pkt.lrank	   = src_lrank;
    pkt.tag	   = tag;
    pkt.len	   = len;
    MPID_DO_HETERO(pkt.msgrep = msgrep);
	
    DEBUG_PRINT_SEND_PKT("S Sending extra-long message",&pkt)

    MPID_PKT_PACK( &pkt, sizeof(MPID_PKT_HEAD_T), dest );

    /* Send as packet only */
    MPID_DRAIN_INCOMING_FOR_TINY(1);
    MPID_SendControl( &pkt, pkt_len, dest );

    /* Send the body of the message */
    MPID_ISendChannel( buf, len, dest, shandle->sid );
    shandle->wait	 = MPID_CH_Eagern_wait_send;
    shandle->test	 = MPID_CH_Eagern_test_send;
/*     shandle->finish	 = 0; */
    shandle->is_complete = 0;

    return MPI_SUCCESS;
}

int MPID_CH_Eagern_cancel_send( shandle )
MPIR_SHANDLE *shandle;
{
    return 0;
}

int MPID_CH_Eagern_test_send( shandle )
MPIR_SHANDLE *shandle;
{
    /* Test for completion */

    if (!shandle->is_complete) {
	if (MPID_TSendChannel( shandle->sid )) {
	    shandle->is_complete = 1;
	    if (shandle->finish) 
		(shandle->finish)( shandle );
	}
    }
    return MPI_SUCCESS;
}

int MPID_CH_Eagern_wait_send( shandle )
MPIR_SHANDLE *shandle;
{
    if (!shandle->is_complete) {
#ifdef MPID_LIMITED_BUFFERS
	/* We do this to keep us from blocking in a wait in the event that
	   we must handle some incoming messages before we can execute the
	   wait. */
	while (!MPID_TestNBSendTransfer(shandle->sid))
	    (void) MPID_DeviceCheck( MPID_NOTBLOCKING );
	/* Once we have it, the message is completed */
#else
	MPID_WSendChannel( shandle->sid );
#endif
	shandle->is_complete = 1;
	if (shandle->finish) 
	    (shandle->finish)( shandle );
	}
    return MPI_SUCCESS;
}
void MPID_CH_Eagern_delete( p )
MPID_Protocol *p;
{
    FREE( p );
}

MPID_Protocol *MPID_CH_Eagern_setup()
{
    MPID_Protocol *p;

    p = (MPID_Protocol *) MALLOC( sizeof(MPID_Protocol) );
    if (!p) return 0;
    p->send	   = MPID_CH_Eagerb_send;
    p->recv	   = MPID_CH_Eagerb_recv;
    p->isend	   = MPID_CH_Eagern_isend;
    p->wait_send   = 0;
    p->push_send   = 0;
    p->cancel_send = MPID_CH_Eagern_cancel_send;
    p->irecv	   = MPID_CH_Eagerb_irecv;
    p->wait_recv   = 0;
    p->push_recv   = 0;
    p->cancel_recv = 0;
    p->do_ack      = 0;
    p->unex        = MPID_CH_Eagerb_save;
    p->delete      = MPID_CH_Eagern_delete;

    return p;
}
